/****************************************************************************/
/*                                                                          */
/*  High level (and hopefully stable) client interface. Allows you to       */
/*  create and report events to the server and look at its response         */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <idsa_internal.h>

static int _idsa_setup(IDSA_CONNECTION * c);
static int _idsa_connect(char *f);
static int _idsa_reply(IDSA_CONNECTION * c);

/****************************************************************************/
/* Does       : set up the entire thing                                     */
/* Parameters : service - your name, credential - NULL for the time being   */
/*              flags - IDSA_F_*                                            */
/* Returns    : pointer to connection structure on success, else NULL       */

IDSA_CONNECTION *idsa_open(char *service, char *credential, int flags)
{
  IDSA_CONNECTION *c;
  int i;
  char *t;
  struct sigaction sag;

  if (!(flags & IDSA_F_SIGPIPE)) {
    sag.sa_handler = SIG_IGN;
    sigfillset(&(sag.sa_mask));
    sag.sa_flags = SA_RESTART;
    sigaction(SIGPIPE, &sag, NULL);
  }
  if (service) {
    c = malloc(sizeof(IDSA_CONNECTION));
    if (c != NULL) {
      c->c_cache = idsa_event_new(0);
      c->c_template = idsa_event_new(0);
      c->c_reply = idsa_event_new(0);
      if ((c->c_cache != NULL) && (c->c_template != NULL) && (c->c_reply != NULL)) {
	strncpy(c->c_service, service, IDSA_M_MEDIUM - 1);
	c->c_service[IDSA_M_MEDIUM - 1] = '\0';

	if (credential) {
	  strncpy(c->c_credential, credential, IDSA_M_MEDIUM);
	  c->c_credential[IDSA_M_MEDIUM - 1] = '\0';
	} else {
	  c->c_credential[0] = '\0';
	}

	c->c_flags = flags;

	c->c_count = 1;
	c->c_result = 0;
	c->c_cache = NULL;

#ifdef WANTS_PROF
	c->c_systime = 0;
	c->c_libtime = 0;
	c->c_waltime = 0;
	c->c_profnum = 0;
#endif

	for (i = 0; i < IDSA_M_RISKS; i++) {
	  c->c_scale[i] = IDSA_RISKSCALE;
	}

	if (c->c_flags & IDSA_F_FAILOPEN) {
	  /* on failure allow all risk ratings */
	  for (i = 0; i < IDSA_M_RISKS; i++) {
	    t = idsa_reserved_namebynumber(IDSA_N_FAILARISK + i);
	    if (t) {
	      c->c_fail[i] = idsa_unit_new(t, IDSA_T_RISK, "[-1.0/1.0]");
	    }
	  }
	} else {
	  /* on failure allow no risk ratings */
	  for (i = 0; i < IDSA_M_RISKS; i++) {
	    c->c_fail[i] = NULL;
	  }
	}

	/* no preselection by default */
	for (i = 0; i < IDSA_M_RISKS; i++) {
	  c->c_pre[i] = NULL;
	}

	idsa_message_default(c->c_template, c->c_service, c->c_service, c->c_service);

	c->c_reason[0] = '\0';

	c->c_backoff = 1;
	c->c_error = _idsa_setup(c);	/* start talking to other side if possible */
      } else {
	if (c->c_template) {
	  idsa_event_free(c->c_template);
	  c->c_template = NULL;
	}
	if (c->c_cache) {
	  idsa_event_free(c->c_cache);
	  c->c_cache = NULL;
	}
	if (c->c_reply) {
	  idsa_event_free(c->c_reply);
	  c->c_reply = NULL;
	}
	free(c);
	c = NULL;
      }
    } else {
      c = NULL;
    }
  } else {
    c = NULL;
  }

  return c;
}

/****************************************************************************/
/* Does       : Deallocate resources associated with connection (close file */
/*              descriptor and release memory)                              */

int idsa_close(IDSA_CONNECTION * c)
{
  int result = 0;
  int i;

  if (c != NULL) {
    /* zap last event */
    if (c->c_cache != NULL) {
      idsa_event_free(c->c_cache);
      c->c_cache = NULL;
    }
    if (c->c_template != NULL) {
      idsa_event_free(c->c_template);
      c->c_template = NULL;
    }
    if (c->c_reply != NULL) {
      idsa_event_free(c->c_reply);
      c->c_reply = NULL;
    }
    /* zap connection */
    result = close(c->c_fd);
    c->c_fd = (-1);

    for (i = 0; i < IDSA_M_RISKS; i++) {
      if (c->c_pre[i]) {
	idsa_unit_free(c->c_pre[i]);
	c->c_pre[i] = NULL;
      }
    }
    for (i = 0; i < IDSA_M_RISKS; i++) {
      if (c->c_fail[i]) {
	idsa_unit_free(c->c_fail[i]);
	c->c_fail[i] = NULL;
      }
    }

    free(c);
  }
  return result;
}

/* single line usage ******************************************************* */

/****************************************************************************/
/* Does       : creates an event from arguments and logs it. The variable   */
/*              argument list is given as a triple of name, type and value  */
/*              where the value is represented as a string. The list is     */
/*              terminated by a null name string                            */

int idsa_scan(IDSA_CONNECTION * c, char *n, char *s, int f, double cr, double ar, double ir,...)
{
  IDSA_EVENT *e;
  va_list ap;
  int result = IDSA_L_DENY;
#ifdef WANTS_PROF
  struct tms start, stop;
  clock_t wall;
#endif

#ifdef WANTS_PROF
  wall = times(&start);
  c->c_profnum++;
#endif

  va_start(ap, ir);

  if (c) {			/* major failure */
    e = idsa_event(c);
    if (e) {
      cr = cr * c->c_scale[0];
      ar = ar * c->c_scale[1];
      ir = ir * c->c_scale[2];

      if (idsa_event_vscan(e, n, s, f, cr, ar, ir, ap)) {
	idsa_free(c, e);
	if (c->c_flags & IDSA_F_FAILOPEN) {
	  result = IDSA_L_ALLOW;
	}
      } else {
	/* log does its own free */
	result = idsa_log(c, e);
      }

    } else {
      if (c->c_flags & IDSA_F_FAILOPEN) {
	result = IDSA_L_ALLOW;
      }
    }
  }
  va_end(ap);

#ifdef WANTS_PROF
  wall = (times(&stop) - wall);
  c->c_waltime = c->c_libtime + wall;
  c->c_libtime = c->c_libtime + stop.tms_utime - start.tms_utime;
  c->c_systime = c->c_systime + stop.tms_stime - start.tms_stime;
#endif

  return result;
}

/****************************************************************************/
/* Does       : creates an event from arguments and logs it. The variable   */
/*              argument list is given as a triple of name, type and value  */
/*              where the value is a pointer. The list is terminated by a   */
/*              null name string                                            */

int idsa_set(IDSA_CONNECTION * c, char *n, char *s, int f, double cr, double ar, double ir,...)
{
  IDSA_EVENT *e;
  va_list ap;
  int result = IDSA_L_DENY;

#ifdef WANTS_PROF
  struct tms start, stop;
  clock_t wall;
#endif

#ifdef WANTS_PROF
  wall = times(&start);
  c->c_profnum++;
#endif

  va_start(ap, ir);

  if (c) {			/* major failure */
    e = idsa_event(c);
    if (e) {
      cr = cr * c->c_scale[0];
      ar = ar * c->c_scale[1];
      ir = ir * c->c_scale[2];

      if (idsa_event_vset(e, n, s, f, cr, ar, ir, ap)) {
	idsa_free(c, e);
	if (c->c_flags & IDSA_F_FAILOPEN) {
	  result = IDSA_L_ALLOW;
	}
      } else {
	/* log does its own free */
	result = idsa_log(c, e);
      }

    } else {
      if (c->c_flags & IDSA_F_FAILOPEN) {
	result = IDSA_L_ALLOW;
      }
    }
  }
  va_end(ap);

#ifdef WANTS_PROF
  wall = (times(&stop) - wall);
  c->c_waltime = c->c_libtime + wall;
  c->c_libtime = c->c_libtime + stop.tms_utime - start.tms_utime;
  c->c_systime = c->c_systime + stop.tms_stime - start.tms_stime;
#endif

  return result;
}

/* event setup ************************************************************* */

/****************************************************************************/
/* Does       : return a new event which can later be logged. The event     */
/*              contains reasonable defaults which can be modified with     */
/*              with idsa_template                                          */
/* Parameters : pointer to event, NULL on failure                           */

IDSA_EVENT *idsa_event(IDSA_CONNECTION * c)
{
  IDSA_EVENT *result;

  if (c) {
    if (c->c_cache == NULL) {
      result = idsa_event_new(0);
    } else {
      result = c->c_cache;
      c->c_cache = NULL;
    }
    if (result) {
      idsa_event_copy(result, c->c_template);
    }
  } else {
    result = NULL;
  }

  return result;
}

/****************************************************************************/
/* Does       : Sets risks scale factor                                     */
/* Returns    : zero on success, nonzero otherwise                          */

int idsa_scale(IDSA_CONNECTION * c, double ar, double cr, double ir)
{
  int result = 0;
  double s[IDSA_M_RISKS];
  int i;

  s[0] = ar;
  s[1] = cr;
  s[2] = ir;

  if (c) {
    for (i = 0; i < IDSA_M_RISKS; i++) {
      if ((s[i] > 1.0) || (s[i] < 0.0)) {
	result++;
      } else {
	c->c_scale[i] = s[i];
      }
    }
  } else {
    result++;
  }

  return result;
}

/****************************************************************************/
/* Does       : Make event e the template for all subsequent calls to       */
/*              idsa_event                                                  */
/* Parameters : e - event to use as template, if NULL revert to original    */

void idsa_template(IDSA_CONNECTION * c, IDSA_EVENT * e)
{
  if (c) {
    if (e) {
      idsa_event_copy(c->c_template, e);
    } else {
      /* idsa_message */
      idsa_message_default(c->c_template, c->c_service, c->c_service, c->c_service);
    }

    if (!(c->c_flags & IDSA_F_KEEP)) {
      idsa_free(c, e);
    }
  }
}

/****************************************************************************/
/* Does       : release event memory, only required if F_KEEP set           */

void idsa_free(IDSA_CONNECTION * c, IDSA_EVENT * e)
{
  if (c) {
    if (c->c_cache == NULL) {
      c->c_cache = e;
    } else {
      idsa_event_free(e);
    }
  } else {
    idsa_event_free(e);
  }
}

/* modify required fields ************************************************** */

/****************************************************************************/
/* Does       : Sets process id of event                                    */

int idsa_pid(IDSA_EVENT * e, pid_t p)
{
  if (idsa_event_setbynumber(e, IDSA_N_PID, &p)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets owner of event                                         */

int idsa_uid(IDSA_EVENT * e, uid_t u)
{
  if (idsa_event_setbynumber(e, IDSA_N_UID, &u)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets group owner of event                                   */

int idsa_gid(IDSA_EVENT * e, gid_t g)
{
  if (idsa_event_setbynumber(e, IDSA_N_GID, &g)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets event timestamp                                        */
/* Notes      : Pointless, log will add its own                             */

int idsa_time(IDSA_EVENT * e, time_t t)
{
  if (idsa_event_setbynumber(e, IDSA_N_TIME, &t)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets hostname for this event                                */

int idsa_host(IDSA_EVENT * e, char *h)
{
  if (idsa_event_scanbynumber(e, IDSA_N_HOST, h)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets name field for event                                   */

int idsa_name(IDSA_EVENT * e, char *n)
{
  if (idsa_event_scanbynumber(e, IDSA_N_NAME, n)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets scheme/namespace for event                             */

int idsa_scheme(IDSA_EVENT * e, char *s)
{
  if (idsa_event_scanbynumber(e, IDSA_N_SCHEME, s)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets service name for event                                 */

int idsa_service(IDSA_EVENT * e, char *s)
{
  if (idsa_event_scanbynumber(e, IDSA_N_SERVICE, s)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets risk rating of event                                   */

int idsa_risks(IDSA_EVENT * e, int f, double a, double c, double i)
{
  if (idsa_event_setbynumber(e, IDSA_N_HONOUR, &f) &&
      idsa_event_setbynumber(e, IDSA_N_ARISK, &a) &&
      idsa_event_setbynumber(e, IDSA_N_CRISK, &c) &&
      idsa_event_setbynumber(e, IDSA_N_IRISK, &i)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Sets flag indicating if a deny will be honoured             */

int idsa_honour(IDSA_EVENT * e, int f)
{
  if (idsa_event_setbynumber(e, IDSA_N_HONOUR, &f)) {
    return 0;
  } else {
    return 1;
  }
}

/* add more fields ********************************************************* */

/****************************************************************************/
/* Does       : Lets you explain your event to the sysadmin                 */

int idsa_comment(IDSA_EVENT * e, char *m,...)
{
  int result = 0;
  char buffer[IDSA_M_MEDIUM];
  va_list ap;
  int pr;

  va_start(ap, m);

  pr = vsnprintf(buffer, IDSA_M_MEDIUM, m, ap);
  if (pr > 0) {
    if (idsa_event_scanappend(e, idsa_reserved_namebynumber(IDSA_N_COMMENT), IDSA_T_STRING, buffer) == NULL) {
      result = 1;
    }
  } else {
    result = 1;
  }

  va_end(ap);

  return result;
}

/****************************************************************************/
/* Does       : Append a string unit to event                               */
/* Returns    : zero on success, nonzero otherwise                          */

int idsa_add_string(IDSA_EVENT * e, char *n, char *s)
{
  if (idsa_event_scanappend(e, n, IDSA_T_STRING, s)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Append an integer unit to event                             */
/* Returns    : zero on success, nonzero otherwise                          */

int idsa_add_integer(IDSA_EVENT * e, char *n, int i)
{
  if (idsa_event_setappend(e, n, IDSA_T_INT, &i)) {
    return 0;
  } else {
    return 1;
  }
}

/****************************************************************************/
/* Does       : Append any type to event (access to more than int or string) */
/* Parameters : n - name of field, t - its type, s - pointer to data        */
/* Returns    : zero on success, nonzero otherwise                          */

int idsa_add_scan(IDSA_EVENT * e, char *n, unsigned int t, char *s)
{
  if (idsa_event_scanappend(e, n, t, s)) {
    return 0;
  } else {
    return 1;
  }
}

int idsa_add_set(IDSA_EVENT * e, char *n, unsigned int t, void *p)
{
  /* IDSA_UNIT *u; */
  if (idsa_event_setappend(e, n, t, p)) {
    /* return idsa_unit_check(u); */
    return 0;
  } else {
    return 1;
  }
}

int idsa_add_unit(IDSA_EVENT * e, IDSA_UNIT * u)
{
  if (idsa_event_setappend(e, u->u_name, u->u_type, u->u_ptr)) {
    return 0;
  } else {
    return 1;
  }
}

/* logging and its results ************************************************* */

/****************************************************************************/
/* Does       : send event off to other side, delete event unless F_KEEP    */
/*              set.                                                        */
/* Returns    : L_OK if request allowed, L_DENY if not                      */

int idsa_log(IDSA_CONNECTION * c, IDSA_EVENT * e)
{
  int result, msg, i, pre;
  IDSA_UNIT *u;

  result = IDSA_L_ALLOW;	/* hope to allow */
  pre = 1;			/* hope to do preselection */

  for (i = 0; i < IDSA_M_RISKS; i++) {
    if (c->c_pre[i]) {
      u = idsa_event_unitbynumber(e, IDSA_N_ARISK + i);
      if (u) {
	switch (idsa_unit_compare(c->c_pre[i], u)) {
	case IDSA_COMPARE_INTERSECT:
	case IDSA_COMPARE_EQUAL:	/* looks good, this risk in range */
	  break;
	default:		/* not in range, have to send */
	  pre = 0;
	  break;
	}
      } else {
	pre = 0;
      }
    } else {
      pre = 0;
    }
  }

  if (pre) {
    snprintf(c->c_reason, IDSA_M_LONG, "preselection filter active");
  } else {			/* pre selection filter did not kick in */


    if ((c->c_error % (c->c_backoff)) == 0) {	/* try to speak to other party */

      idsa_time(e, time(NULL));

      if (c->c_error == 0) {	/* no previous errors, attempt a normal write */
	c->c_length = 0;
	do {
	  msg = idsa_event_write(c->c_fd, e, &(c->c_length));
	} while (msg == IDSA_IO_WAIT);
#ifdef TRACE
	fprintf(stderr, "idsa_log(): wrote message, code <%d>\n", msg);
#endif
      } else {			/* previous errors, don't bother writing until restart */
	msg = IDSA_IO_FAIL;
      }

      if (msg == IDSA_IO_FAIL) {	/* failure, try to reconnect */

#ifdef TRACE
	fprintf(stderr, "idsa_log(): write failed, attempting to reconnect\n");
#endif

	/* pessimist, bump backoff */
	c->c_error = 1;
	c->c_backoff = c->c_backoff + 1;

	if (c->c_fd != (-1)) {
	  close(c->c_fd);
	  c->c_fd = (-1);
	}
	if (_idsa_setup(c) == 0) {	/* good, could reconnect */

	  c->c_length = 0;
	  do {
	    msg = idsa_event_write(c->c_fd, e, &(c->c_length));
	  } while (msg == IDSA_IO_WAIT);

	  if (msg == IDSA_IO_OK) {	/* even better, could write so clear error */
#ifdef TRACE
	    fprintf(stderr, "idsa_log(): reconnect succeeded\n");
#endif
	    c->c_error = 0;
	    c->c_backoff = 1;
	  }
	}
      }
      /* if IO_FAIL then error>0 */
      /* if IO_OK then error=0 */

      if (msg == IDSA_IO_OK) {

	c->c_length = 0;
	do {
	  msg = idsa_event_read(c->c_fd, c->c_reply, &(c->c_length));
	} while (msg == IDSA_IO_WAIT);

#ifdef TRACE
	fprintf(stderr, "idsa_log(): read reply, code <%d>\n", msg);
#endif

	if (msg == IDSA_IO_OK) {
	  if (idsa_reply_check(c->c_reply) == 0) {
	    _idsa_reply(c);
	    result = c->c_reply->e_result;
	  } else {
	    c->c_error = c->c_error + 1;
	  }
	} else {
	  c->c_error = c->c_error + 1;
	}
      }
    } else {			/* previous failures and backoff kicked in, so just wait */
#ifdef TRACE
      fprintf(stderr, "idsa_log(): backoff active, %d/%d\n", c->c_error, c->c_backoff);
#endif
      c->c_error = c->c_error + 1;
    }

    if (c->c_error) {		/* on failure try to see if fail filter active */
#ifdef TRACE
      fprintf(stderr, "idsa_log(): failure, checking fail filter\n");
#endif
      result = IDSA_L_ALLOW;	/* hope to allow */
      for (i = 0; i < IDSA_M_RISKS; i++) {
	if (c->c_fail[i]) {
	  u = idsa_event_unitbynumber(e, IDSA_N_ARISK + i);
	  if (u) {
	    switch (idsa_unit_compare(c->c_fail[i], u)) {
	    case IDSA_COMPARE_INTERSECT:
	    case IDSA_COMPARE_EQUAL:	/* looks good, this risk in range */
	      break;
	    default:		/* not in range, have to deny */
	      result = IDSA_L_DENY;
	      break;
	    }
	  } else {
	    result = IDSA_L_DENY;
	  }
	} else {
	  result = IDSA_L_DENY;
	}
      }
    }				/* end of fail filter */
  }

  if (!(c->c_flags & IDSA_F_KEEP)) {
    idsa_free(c, e);
  }
  /* mask out any other return codes */
  return (result & IDSA_L_DENY);
}

/****************************************************************************/
/* Does       : Returns nonzero if an error has occured, generally little   */
/*              reason to look at it                                        */

int idsa_error(IDSA_CONNECTION * c)
{
  return c->c_error;
}

/****************************************************************************/
/* Does       : Returns a string containing an reason for the last          */
/*              L_* code of idsa_log, if available, otherwise NULL          */

char *idsa_reason(IDSA_CONNECTION * c)
{
  if (c->c_reason[0] == '\0') {
    return NULL;
  }
  return c->c_reason;
}

/* internal functions ****************************************************** */

static int _idsa_setup(IDSA_CONNECTION * c)
{
  int result;
  char *f;

  if (c->c_flags & IDSA_F_ENV) {
    /* I wonder if this is wise - setuid should never use this option */
    /* so maybe nanny test if effective uid != real uid ? */
    f = getenv("IDSA_SOCKET");
  } else {
    f = NULL;
  }

  c->c_length = 0;
  c->c_fd = _idsa_connect(f ? f : IDSA_SOCKET);
  if (c->c_fd == (-1)) {
    result = 1;
  } else {
    fcntl(c->c_fd, F_SETFD, 1);	/* only one at a time please */
    result = 0;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_setup(): setup %s\n", result ? "failed" : "ok");
#endif

  return result;
}

/****************************************************************************/
/* Does       : processes units sent in reply                               */
/* Notes      : more could be done here - upload fancy filter expressions   */

static int _idsa_reply(IDSA_CONNECTION * c)
{
  int result = 0;
  unsigned int i, n;
  IDSA_EVENT *e;
  IDSA_UNIT *u;

  c->c_reason[0] = '\0';

  e = c->c_reply;
  for (i = 0; i < e->e_count; i++) {
    u = idsa_event_unitbynumber(e, i);
    if (u) {
      n = idsa_reserved_numberbyname(u->u_name);
      if (u->u_type == idsa_reserved_typebynumber(n)) {
	switch (n) {
	case IDSA_N_REASON:
	  idsa_unit_print(u, c->c_reason, IDSA_M_MEDIUM - 1);
	  c->c_reason[IDSA_M_MEDIUM] = '\0';
	  break;
	case IDSA_N_PREARISK:
	case IDSA_N_PRECRISK:
	case IDSA_N_PREIRISK:
#ifdef TRACE
	  fprintf(stderr, "_idsa_reply(): setting prefilter[%d]\n", n - IDSA_N_PREARISK);
#endif
	  if (c->c_pre[n - IDSA_N_PREARISK]) {
	    idsa_unit_copy(c->c_pre[n - IDSA_N_PREARISK], u);
	  } else {
	    c->c_pre[n - IDSA_N_PREARISK] = idsa_unit_dup(u);
	  }
	  break;
	case IDSA_N_FAILARISK:
	case IDSA_N_FAILCRISK:
	case IDSA_N_FAILIRISK:
#ifdef TRACE
	  fprintf(stderr, "_idsa_reply(): setting fail filter[%d]\n", n - IDSA_N_FAILARISK);
#endif
	  if (c->c_fail[n - IDSA_N_FAILARISK]) {
	    idsa_unit_copy(c->c_fail[n - IDSA_N_FAILARISK], u);
	  } else {
	    c->c_fail[n - IDSA_N_FAILARISK] = idsa_unit_dup(u);
	  }
	  break;
	  /* default : is to ignore */
	}
      }
    }
  }

  return result;
}

static int _idsa_connect(char *f)
{
  struct sockaddr_un addr;
  int result;

  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, f, sizeof(addr.sun_path));

  result = socket(AF_UNIX, SOCK_STREAM, 0);
  if (result != (-1)) {
    if (connect(result, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
      close(result);
      result = (-1);
    }
  }
  return result;
}

#ifdef WANTS_PROF
int idsa_prof(IDSA_CONNECTION * c, FILE * fp)
{

  if (c->c_profnum > 0) {
    fprintf(fp, "lib %ld/%d=%ld, system %ld/%d=%ld, wall %ld/%d=%ld\n", c->c_libtime, c->c_profnum, c->c_libtime / c->c_profnum, c->c_systime, c->c_profnum, c->c_systime / c->c_profnum, c->c_waltime, c->c_profnum, c->c_waltime / c->c_profnum);
  } else {
    fprintf(fp, "no events\n");
  }

  return 0;
}
#endif
