
/****************************************************************************/
/*                                                                          */
/*  Functions to scan (parse), print (display in human readable format),    */
/*  set (modify in dangerous ways), check integrity and compare various     */
/*  data types. At some stage this should be improved to have proper IO     */
/*  functions for each unit - instead of the half-baked scan and print.     */
/*  Problem is that types which use symbolic names are not chroot()         */
/*  safe - they require support files in jail.                              */
/*                                                                          */
/*  Warning - many of these functions are incomplete and just borrow the    */
/*  integer routines. So before porting to any arch where time_t != int     */
/*  the type handlers will have to implemented.                             */
/*                                                                          */
/****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <idsa_internal.h>

/* string handler ******************************************************* */

int _idsa_string_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;

  result = strncmp(a->u_ptr, b->u_ptr, (IDSA_S_MASK & a->u_type));

  if (result < 0) {
    result = IDSA_COMPARE_LESS;
  } else if (result > 0) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

#ifdef TRACE
  fprintf(stderr, "compare_string(): comparing of <%s> against <%s>\n", a->u_ptr, b->u_ptr);
#endif

  return result;
}

int _idsa_string_print(IDSA_UNIT * u, char *s, int l)
{
  int i;

  u->u_ptr[(u->u_type & IDSA_S_MASK) - 1] = '\0';
  i = strlen(u->u_ptr);
  if (l < i) {
    memcpy(s, u->u_ptr, l);	/* feeble best effort */
    return -1;
  } else {
    memcpy(s, u->u_ptr, i);
    return i;
  }
}

int _idsa_string_scan(IDSA_UNIT * u, char *s)
{
  int i;
  int result = 0;

  i = strlen(s);
  if (i > ((u->u_type & IDSA_S_MASK) - 1)) {
    i = (u->u_type & IDSA_S_MASK) - 1;
    result = 1;			/* string too long, but we try our best */
  }
  strncpy(u->u_ptr, s, i);
  u->u_ptr[i] = '\0';

  return result;
}

int _idsa_string_set(IDSA_UNIT * u, void *p)
{
  return _idsa_string_scan(u, p);
}

int _idsa_string_check(IDSA_UNIT * u)
{
  u->u_ptr[(u->u_type & IDSA_S_MASK) - 1] = '\0';
  return 0;
}

int _idsa_string_display(IDSA_UNIT * u, char *s, int l)
{
  return _idsa_string_print(u, s, l);
}

/* integer handler ****************************************************** */

int _idsa_int_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  unsigned int x, y;
  int result;

  memcpy(&x, a->u_ptr, sizeof(int));
  memcpy(&y, b->u_ptr, sizeof(int));

  if (x < y) {
    result = IDSA_COMPARE_LESS;
  } else if (x > y) {
    result = IDSA_COMPARE_MORE;
  } else {
    result = IDSA_COMPARE_EQUAL;
  }

  return result;
}

int _idsa_int_print(IDSA_UNIT * u, char *s, int l)
{
  unsigned int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  return snprintf(s, l, "%u", i);
}

int _idsa_int_scan(IDSA_UNIT * u, char *s)
{
  unsigned int i;

  if (isdigit(s[0])) {
    i = atoi(s);
    memcpy(u->u_ptr, &i, sizeof(int));
    return 0;
  } else {
    return -1;
  }

}

int _idsa_int_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(int));

  return 0;
}

/* no _idsa_int_check */

int _idsa_int_display(IDSA_UNIT * u, char *s, int l)
{
  return _idsa_int_print(u, s, l);
}

/* uid lookup: FIXME: assumes that sizeof(uid_t)==sizeof(int) *********** */

int _idsa_uid_scan(IDSA_UNIT * u, char *s)
{
  struct passwd *pw;

  if (isdigit(s[0])) {
    return _idsa_int_scan(u, s);
  } else {
    pw = getpwnam(s);
    if (pw) {
      memcpy(u->u_ptr, &(pw->pw_uid), sizeof(uid_t));
      return 0;
    }
  }

  return 1;
}

int _idsa_uid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(uid_t));

  return 0;
}

/* gid lookup: FIXME: assumes that sizeof(gid_t)==sizeof(int) *********** */

int _idsa_gid_scan(IDSA_UNIT * u, char *s)
{
  struct group *gr;

  if (isdigit(s[0])) {
    return _idsa_int_scan(u, s);
  } else {
    gr = getgrnam(s);
    if (gr) {
      memcpy(u->u_ptr, &(gr->gr_gid), sizeof(gid_t));
      return 0;
    }
  }

  return 1;
}

int _idsa_gid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(gid_t));

  return 0;
}

/* pid lookup: FIXME: assumes that sizeof(pid_t)==sizeof(int) *********** */

int _idsa_pid_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(pid_t));

  return 0;
}

/* time handler (to be done properly) *********************************** */

int _idsa_time_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(time_t));

  return 0;
}

/* flag handler (nice and simple) *************************************** */

int _idsa_flag_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  unsigned int x, y;

  memcpy(&x, a->u_ptr, sizeof(int));
  memcpy(&y, b->u_ptr, sizeof(int));

  if (x) {
    if (y) {
      return IDSA_COMPARE_EQUAL;
    } else {
      return IDSA_COMPARE_LESS;
    }
  } else {
    if (y) {
      return IDSA_COMPARE_MORE;
    } else {
      return IDSA_COMPARE_EQUAL;
    }
  }
}

int _idsa_flag_print(IDSA_UNIT * u, char *s, int l)
{
  int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  if (l > 0) {
    if (i) {
      s[0] = '1';
    } else {
      s[0] = '0';
    }
    return 1;
  } else {
    return -1;
  }
}

int _idsa_flag_scan(IDSA_UNIT * u, char *s)
{
  unsigned int i;

  if (s[0] == '0' || s[0] == 'f' || s[0] == 'F') {
    i = 0;
  } else {
    i = 1;
  }
  memcpy(u->u_ptr, &i, sizeof(int));

  return 0;
}

int _idsa_flag_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, sizeof(int));

  return 0;
}

int _idsa_flag_display(IDSA_UNIT * u, char *s, int l)
{
  int i;

  memcpy(&i, u->u_ptr, sizeof(int));

  return snprintf(s, l, "%s", i ? "true" : "false");
}

/* risk handler (not simple) ******************************************** */

int _idsa_risk_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;

  unsigned short ab, ae, bb, be;

  memcpy(&ab, a->u_ptr, sizeof(short));
  memcpy(&ae, a->u_ptr + sizeof(short), sizeof(short));

  memcpy(&bb, b->u_ptr, sizeof(short));
  memcpy(&be, b->u_ptr + sizeof(short), sizeof(short));

  if (ae < bb) {
    result = IDSA_COMPARE_LESS;
  } else if (ab > be) {
    result = IDSA_COMPARE_MORE;
  } else if ((ab == bb) && (ae == be)) {
    result = IDSA_COMPARE_EQUAL;
  } else {
    result = IDSA_COMPARE_INTERSECT;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_compare(): [%d,%d] ? [%d,%d] yields %d\n", ab, ae, bb, be, result);
#endif

  return result;
}

int _idsa_risk_print(IDSA_UNIT * u, char *s, int l)
{
  unsigned short r;
  memcpy(&r, u->u_ptr, sizeof(short));

  return snprintf(s, l, "%.3f", ((2.0 * (double) r) / (double) (IDSA_M_RISKTOP)) - 1.0);
}

int _idsa_risk_scan(IDSA_UNIT * u, char *s)
{
  int i;
  double a, b;
  unsigned short v[2];

  i = 0;

  if (isdigit(s[i]) || (s[i] == '-')) {
    a = (1.0 + atof(s + i)) / 2.0;
    b = a;
  } else {
    if (s[i] == '\0')
      return 1;

    i++;
    a = (1.0 + atof(s + i)) / 2.0;
    while ((s[i] != '\0') && (s[i] != '/') && (s[i] != ','))
      i++;
    if (s[i] == '\0')
      return 1;

    i++;
    b = (1.0 + atof(s + i)) / 2.0;
    while ((s[i] != '\0') && (s[i] != '>') && (s[i] != ']'))
      i++;
    if (s[i] == '\0')
      return 1;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_read(): floating %.2f-%.2f\n", a, b);
#endif

  if ((a < -0.00001) || (a > 1.00001) || (b < -0.00001) || (b > 1.00001))
    return 1;

  if (a > b) {
    v[0] = (unsigned short) (b * IDSA_M_RISKTOP);
    if (s[i] == '>')
      v[0] += 1;
    v[1] = (unsigned short) (a * IDSA_M_RISKTOP);
    if (s[0] == '<')
      v[1] -= 1;
  } else {
    v[0] = (unsigned short) (a * IDSA_M_RISKTOP);
    if (s[0] == '<')
      v[0] += 1;
    v[1] = (unsigned short) (b * IDSA_M_RISKTOP);
    if (s[i] == '>')
      v[1] -= 1;
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_read(): [%04x-%04x]\n", v[0], v[1]);
#endif

  memcpy(u->u_ptr, v, 2 * sizeof(short));

  return 0;
}

int _idsa_risk_check(IDSA_UNIT * u)
{				/* unused */
  return 0;
}

int _idsa_risk_set(IDSA_UNIT * u, void *p)
{
  double d;
  unsigned short v[2];

  d = *((double *) p);

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_set(): got %.3f, change to %.3f\n", d, (d + 1.0) / 2.0);
#endif
  d = (d + 1.0) / 2.0;

  if ((d < -0.00001) || (d > 1.00001))
    return 1;

  v[0] = (unsigned short) (d * IDSA_M_RISKTOP);
  v[1] = (unsigned short) (d * IDSA_M_RISKTOP);

#ifdef TRACE
  fprintf(stderr, "_idsa_risk_set(): [%04x-%04x]\n", v[0], v[1]);
#endif

  memcpy(u->u_ptr, v, 2 * sizeof(short));

  return 0;
}

int _idsa_risk_display(IDSA_UNIT * u, char *s, int l)
{
  return _idsa_risk_print(u, s, l);
}

/* errno handler ******************************************************** */

int _idsa_errno_display(IDSA_UNIT * u, char *s, int l)
{
  int i, n;
  char *p;

  memcpy(&i, u->u_ptr, sizeof(int));
  p = strerror(i);

#ifdef TRACE
  fprintf(stderr, "_idsa_errno_display(): errno is <%d>, string <%s>\n", i, p);
#endif

  n = strlen(p);

  if (n > l) {
    memcpy(s, p, l);		/* best effort */
    return -1;
  } else {
    memcpy(s, p, n);
    return n;
  }
}

/* hostname handler (host.some.domain), stored in reverse order to ****** */
/* make matching of .domains easier ************************************* */

int _idsa_host_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  int i;

  i = 0;
  while ((i < (IDSA_S_MASK & a->u_type)) && (a->u_ptr[i] == b->u_ptr[i]) && (a->u_ptr[i] != '\0')) {
    i++;
  }

  if (a->u_ptr[i] == b->u_ptr[i]) {
    result = IDSA_COMPARE_EQUAL;
  } else {
    if (a->u_ptr[i] < b->u_ptr[i]) {
      result = IDSA_COMPARE_MORE;
    } else {			/* a > b */
      result = IDSA_COMPARE_LESS;
    }
    if (i > 0) {		/* check for trailing . for network match */
      if ((a->u_ptr[i] == '\0') || (b->u_ptr[i] == '\0')) {
	if ((b->u_ptr[i - 1] == '.') && (a->u_ptr[i - 1] == b->u_ptr[i - 1])) {
	  result = IDSA_COMPARE_INTERSECT;
	}
      }
    }
  }

#ifdef TRACE
  fprintf(stderr, "compare_host(): compare of <%s> against <%s> yields <%d>\n", a->u_ptr, b->u_ptr, result);
#endif

  return result;
}

int _idsa_host_print(IDSA_UNIT * u, char *s, int l)
{
  int i, j, result;

  i = 0;
  while ((i < (u->u_type & IDSA_S_MASK)) && (u->u_ptr[i] != '\0')) {
    i++;
  }

  if (l < i) {
    result = (-1);
    i = l;
  } else {
    result = i;
  }
#ifdef TRACE
  fprintf(stderr, "print_host(): <%s:%d>\n", u->u_ptr, i);
#endif
  i--;

  j = 0;
  while (i >= 0) {
    s[j] = u->u_ptr[i];
    j++;
    i--;
  }

  return result;
}

int _idsa_host_scan(IDSA_UNIT * u, char *s)
{
  int i, j;
  int result = 0;

  i = strlen(s);
  if (i > ((u->u_type & IDSA_S_MASK) - 1)) {
    i = (u->u_type & IDSA_S_MASK) - 1;
    result = 1;
  }
  u->u_ptr[i] = '\0';
  i--;

  j = 0;
  while (i >= 0) {
    u->u_ptr[j] = s[i];
    j++;
    i--;
  }

  return result;
}

int _idsa_host_set(IDSA_UNIT * u, void *p)
{
  return _idsa_host_scan(u, p);
}

int _idsa_host_check(IDSA_UNIT * u)
{
  u->u_ptr[(u->u_type & IDSA_S_MASK) - 1] = '\0';
  return 0;
}

int _idsa_host_display(IDSA_UNIT * u, char *s, int l)
{
  return _idsa_host_print(u, s, l);
}

/* inet address handler (ip4). Internal storage representation is an **** */
/* array of 2 integers, one for address, and one for (32-mask) ********** */

int _idsa_addr_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  unsigned long int p[2], q[2], m;

  memcpy(p, a->u_ptr, 2 * sizeof(long int));
  memcpy(q, b->u_ptr, 2 * sizeof(long int));
  if (p[1] > q[1]) {
    m = (0xffffffff << (p[1]));
  } else {
    m = (0xffffffff << (q[1]));
  }

  p[0] = (p[0] & m);
  q[0] = (q[0] & m);

  if (p[0] > q[0]) {
    result = IDSA_COMPARE_LESS;
  } else if (p[0] < q[0]) {
    result = IDSA_COMPARE_LESS;
  } else {
    if (p[1] == q[1]) {
      result = IDSA_COMPARE_EQUAL;
    } else {
      result = IDSA_COMPARE_INTERSECT;
    }
  }

#ifdef TRACE
  fprintf(stderr, "_idsa_addr_compare(): hostmask is <0x%08lx> bits, <0x%08lx?0x%08lx>, result <%d>\n", m, p[0], q[0], result);
#endif

  return result;
}

int _idsa_addr_print(IDSA_UNIT * u, char *s, int l)
{
  unsigned long int a[2];
  struct in_addr d;
  int result;

  memcpy(a, u->u_ptr, 2 * sizeof(long int));
  d.s_addr = htonl(a[0]);

  if (a[1] > 0) {
    return snprintf(s, l, "%s/%ld", inet_ntoa(d), 32 - a[1]);
  } else {
    return snprintf(s, l, "%s", inet_ntoa(d));
  }

  return result;
}

int _idsa_addr_scan(IDSA_UNIT * u, char *s)
{
  struct in_addr d;
  int result;
  int i;
  unsigned long int a[2];
  char tmp[16];

  a[0] = 0;
  a[1] = 0;

  i = 0;
  while ((s[i] != '\0') && (s[i] != '/')) {
    i++;
  }

  result = 0;
  if (s[i] == '/') {
    if (i > 15) {
      i = 15;
    }
    memcpy(tmp, s, i);
    tmp[i] = '\0';
    if (inet_aton(tmp, &d) == 0) {
      result++;
    }
    a[1] = 32 - atoi(s + i + 1);
#ifdef TRACE
    fprintf(stderr, "_idsa_addr_scan(): address component is <%s>, number %ld\n", tmp, a[1]);
#endif
    if (a[1] > 32) {
      result++;
    }
  } else {			/* no /bits component, just plain IP address */
    if (inet_aton(s, &d) == 0) {
      result++;
    }
  }
  a[0] = ntohl(d.s_addr);

#ifdef TRACE
  fprintf(stderr, "_idsa_addr_scan(): 0x%08lx/%ld\n", a[0], a[1]);
#endif

  memcpy(u->u_ptr, a, 2 * sizeof(long int));

  return result;
}

int _idsa_addr_set(IDSA_UNIT * u, void *p)
{
  unsigned long int a[2];
  memcpy(a, p, sizeof(int));
  a[1] = 0;

#ifdef TRACE
  fprintf(stderr, "_idsa_addr_scan(): 0x%08lx/%ld\n", a[0], a[1]);
#endif

  memcpy(u->u_ptr, a, 2 * sizeof(long int));

  return 0;
}

int _idsa_addr_check(IDSA_UNIT * u)
{
  unsigned long int a[2];
  int result = 0;

  memcpy(a, u->u_ptr, 2 * sizeof(long int));
  if (a[1] > 32) {
    a[1] = 0;
    result++;
  }
  return result;
}

int _idsa_addr_display(IDSA_UNIT * u, char *s, int l)
{
  return _idsa_addr_print(u, s, l);
}

/* port handler (protocol/port) ***************************************** */

int _idsa_port_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int x[2], y[2];
  int result;

  memcpy(&x, a->u_ptr, 2 * sizeof(int));
  memcpy(&y, b->u_ptr, 2 * sizeof(int));

  if (x[0] < y[0]) {
    result = IDSA_COMPARE_LESS;
  } else if (x[0] > y[0]) {
    result = IDSA_COMPARE_MORE;
  } else {
    if (x[1] < y[1]) {
      result = IDSA_COMPARE_LESS;
    } else if (x[1] > y[1]) {
      result = IDSA_COMPARE_MORE;
    } else {
      result = IDSA_COMPARE_EQUAL;
    }
  }

  return result;
}

int _idsa_port_print(IDSA_UNIT * u, char *s, int l)
{
  int x[2];
  memcpy(&x, u->u_ptr, 2 * sizeof(int));

  return snprintf(s, l, "%d/%d", x[0], x[1]);
}

int _idsa_port_scan(IDSA_UNIT * u, char *s)
{
  int x[2];
  int result = 0;
  char p[32];
  int i;
  struct protoent *pe;
  struct servent *se;

  for (i = 0; (s[i] != '\0') && (s[i] != '/'); i++);

  if ((i < 31) && (s[i] == '/')) {
    memcpy(p, s, i);
    p[i] = '\0';
    i++;

    if (isdigit(p[0])) {
      x[0] = atoi(p);
    } else {
      pe = getprotobyname(p);
      if (pe) {
	x[0] = pe->p_proto;
      } else {
	x[0] = 0;
	result++;
      }
    }

    if (isdigit(s[0])) {
      x[1] = atoi(s + i);
    } else {
      se = getservbyname(s + i, p);
      if (se) {
	x[1] = ntohs(se->s_port);
      } else {
	x[1] = 0;
	result++;
      }
    }

#ifdef TRACE
    fprintf(stderr, "_idsa_port_scan(): <%s> maps to %d/%d\n", s, x[0], x[1]);
#endif
    memcpy(u->u_ptr, x, 2 * sizeof(int));

  } else {
    result++;
  }

  return result;
}

int _idsa_port_set(IDSA_UNIT * u, void *p)
{
  memcpy(u->u_ptr, p, 2 * sizeof(int));

  return 0;
}

int _idsa_port_display(IDSA_UNIT * u, char *s, int l)
{
  /* FIXME: should do symbolic names */
  return _idsa_port_print(u, s, l);
}

int _idsa_file_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  int result;
  int i;

  i = 0;
  while ((i < (IDSA_S_MASK & a->u_type)) && (a->u_ptr[i] == b->u_ptr[i]) && (a->u_ptr[i] != '\0')) {
    i++;
  }

  if ((a->u_ptr[i] == '\0') || (b->u_ptr[i] == '\0')) {
    if (a->u_ptr[i] == b->u_ptr[i]) {
      result = IDSA_COMPARE_EQUAL;
    } else {
      if (a->u_ptr[i] == '\0') {
	if ((i > 0) && (a->u_ptr[i - 1] == '/')) {
	  result = IDSA_COMPARE_INTERSECT;
	} else {
	  result = IDSA_COMPARE_LESS;
	}
      } else {
	if ((i > 0) && (b->u_ptr[i - 1] == '/')) {
	  result = IDSA_COMPARE_INTERSECT;
	} else {
	  result = IDSA_COMPARE_MORE;
	}
      }
    }
  } else {
    if (a->u_ptr[i] < b->u_ptr[i]) {
      result = IDSA_COMPARE_LESS;
    } else {
      result = IDSA_COMPARE_MORE;
    }
  }

#ifdef TRACE
  fprintf(stderr, "compare_file(): comparing of <%s> against <%s>\n", a->u_ptr, b->u_ptr);
#endif

  return result;
}

/****************************************************************************/

struct idsa_lookup_unit {
  unsigned int l_type;
  IDSA_FUNCTION_COMPARE l_compare;
  IDSA_FUNCTION_PRINT l_print;
  IDSA_FUNCTION_SCAN l_scan;
  IDSA_FUNCTION_SET l_set;
  IDSA_FUNCTION_CHECK l_check;
  IDSA_FUNCTION_DISPLAY l_display;
  char l_name[IDSA_M_NAME];
};

IDSA_LOOKUP_UNIT _idsa_unit_lookup_table[IDSA_M_TYPES] =
{
  [0] =
  {IDSA_T_NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   ""},
  [IDSA_T_STRING >> IDSA_S_SHIFT] =
  {IDSA_T_STRING,
   &_idsa_string_compare,
   &_idsa_string_print,
   &_idsa_string_scan,
   &_idsa_string_set,
   &_idsa_string_check,
   &_idsa_string_display,
   "string"},
  [IDSA_T_INT >> IDSA_S_SHIFT] =
  {IDSA_T_INT,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_int_scan,
   &_idsa_int_set,
   NULL,
   &_idsa_int_display,
   "integer"},
  [IDSA_T_UID >> IDSA_S_SHIFT] =
  {IDSA_T_UID,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_uid_scan,
   &_idsa_uid_set,
   NULL,
   &_idsa_int_display,
   "uid"},
  [IDSA_T_GID >> IDSA_S_SHIFT] =
  {IDSA_T_GID,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_gid_scan,
   &_idsa_gid_set,
   NULL,
   &_idsa_int_display,
   "gid"},
  [IDSA_T_PID >> IDSA_S_SHIFT] =
  {IDSA_T_PID,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_int_scan,
   &_idsa_pid_set,
   NULL,
   &_idsa_int_display,
   "pid"},
  [IDSA_T_TIME >> IDSA_S_SHIFT] =
  {IDSA_T_TIME,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_int_scan,		/* needs improvement */
   &_idsa_time_set,
   NULL,
   &_idsa_int_display,
   "time"},
  [IDSA_T_FLAG >> IDSA_S_SHIFT] =
  {IDSA_T_FLAG,
   &_idsa_flag_compare,
   &_idsa_flag_print,
   &_idsa_flag_scan,
   &_idsa_flag_set,
   NULL,
   &_idsa_flag_display,
   "flag"},
  [IDSA_T_RISK >> IDSA_S_SHIFT] =
  {IDSA_T_RISK,
   &_idsa_risk_compare,
   &_idsa_risk_print,
   &_idsa_risk_scan,
   &_idsa_risk_set,
   NULL,
   &_idsa_risk_display,
   "risk"},
  [IDSA_T_ERRNO >> IDSA_S_SHIFT] =
  {IDSA_T_ERRNO,
   &_idsa_int_compare,
   &_idsa_int_print,
   &_idsa_int_scan,
   &_idsa_int_set,
   NULL,
   &_idsa_errno_display,
   "errno"},
  [IDSA_T_HOST >> IDSA_S_SHIFT] =
  {IDSA_T_HOST,
   &_idsa_host_compare,
   &_idsa_host_print,
   &_idsa_host_scan,
   &_idsa_host_set,
   &_idsa_host_check,
   &_idsa_host_display,
   "host"},
  [IDSA_T_ADDR >> IDSA_S_SHIFT] =
  {IDSA_T_ADDR,
   &_idsa_addr_compare,
   &_idsa_addr_print,
   &_idsa_addr_scan,
   &_idsa_addr_set,
   &_idsa_addr_check,
   &_idsa_addr_display,
   "addr"},
  {IDSA_T_PORT,
   &_idsa_port_compare,
   &_idsa_port_print,
   &_idsa_port_scan,
   &_idsa_port_set,
   NULL,
   &_idsa_port_display,
   "port"},
  [IDSA_T_FILE >> IDSA_S_SHIFT] =
  {IDSA_T_FILE,
   &_idsa_file_compare,
   &_idsa_string_print,
   &_idsa_string_scan,
   &_idsa_string_set,
   &_idsa_string_check,
   &_idsa_string_display,
   "file"}
};

IDSA_LOOKUP_UNIT *idsa_unit_lookup(unsigned int t)
{
  IDSA_LOOKUP_UNIT *result;
  unsigned int i;

  i = (t >> IDSA_S_SHIFT);
  if (i > 0 && i < IDSA_M_TYPES) {
    result = &(_idsa_unit_lookup_table[i]);
    if (result->l_type != t) {
#ifdef TRACE
      fprintf(stderr, "idsa_unit_lookup(): size mismatch - have <0x%04x>, want <0x%04x>\n", (result->l_type) & IDSA_S_MASK, t & IDSA_S_MASK);
#endif
      result = NULL;
    }
  } else {
#ifdef TRACE
    fprintf(stderr, "idsa_unit_lookup(): unknown type 0x%04x\n", t);
#endif
    result = NULL;

  }

  return result;
}

int idsa_unit_compare(IDSA_UNIT * a, IDSA_UNIT * b)
{
  IDSA_LOOKUP_UNIT *l;

  l = idsa_unit_lookup(a->u_type);
  if (l) {
    return (*(l->l_compare)) (a, b);
  } else {
    return IDSA_COMPARE_LESS;
  }
}

int idsa_unit_print(IDSA_UNIT * u, char *s, int l)
{
  IDSA_LOOKUP_UNIT *t;

  t = idsa_unit_lookup(u->u_type);
  if (t) {
    return (*(t->l_print)) (u, s, l);
  } else {
    return -1;
  }
}

int idsa_unit_scan(IDSA_UNIT * u, char *s)
{
  IDSA_LOOKUP_UNIT *t;

  t = idsa_unit_lookup(u->u_type);
  if (t) {
    return (*(t->l_scan)) (u, s);
  } else {
    return 1;
  }
}

int idsa_unit_set(IDSA_UNIT * u, void *p)
{
  IDSA_LOOKUP_UNIT *t;

#ifdef TRACE
  fprintf(stderr, "idsa_unit_set(): setting type <0x%04x> (address %p)\n", u->u_type, p);
#endif

  t = idsa_unit_lookup(u->u_type);
  if (t) {
    return (*(t->l_set)) (u, p);
  } else {
    return 1;
  }
}

int idsa_unit_check(IDSA_UNIT * u)
{
  IDSA_LOOKUP_UNIT *t;

  t = idsa_unit_lookup(u->u_type);
  if (t) {
    if (t->l_check) {
      return (*(t->l_check)) (u);
    } else {
      return 0;
    }
  } else {
    return 1;
  }
}

int idsa_unit_display(IDSA_UNIT * u, char *s, int l)
{
  IDSA_LOOKUP_UNIT *t;

  t = idsa_unit_lookup(u->u_type);
  if (t) {
    return (*(t->l_display)) (u, s, l);
  } else {
    return -1;
  }
}

char *idsa_unit_name(unsigned int t)
{
  IDSA_LOOKUP_UNIT *l;

  l = idsa_unit_lookup(t);
  if (l) {
    return l->l_name;
  } else {
    return NULL;
  }
}

unsigned int idsa_unit_type(char *n)
{
  int i;

  for (i = 1; i < IDSA_M_TYPES; i++) {
#ifdef TRACE
    fprintf(stderr, "idsa_unit_name(): looking at <%s:0x%04x>\n", _idsa_unit_lookup_table[i].l_name, _idsa_unit_lookup_table[i].l_type);
#endif
    if (strcmp(n, _idsa_unit_lookup_table[i].l_name) == 0) {
      return _idsa_unit_lookup_table[i].l_type;
    }
  }
  return IDSA_T_NULL;
}
