/****************************************************************************/
/*                                                                          */
/*  This used to be a fancy protocol, but got lobotomized and should get    */
/*  pruned down to a single message / reply pair. The protocol itself could */
/*  also do with some improving - possible improvements:                    */
/*                                                                          */
/*    variable sized units (ick, need to hack event.c and unit.c)           */
/*    network ordering                                                      */
/*    plain text                                                            */
/*    some lightweight markup: I am reluctant to use XML because of size    */
/*    and security concerns (matching \0 becomes nontrivial)                */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <idsa_internal.h>

/****************************************************************************/
/* Does       : drop event into buffer                                      */
/* Returns    : amount copied on success, zero on failure                   */

int idsa_event_tobuffer(IDSA_EVENT * e, char *s, int l)
{
  if (l >= e->e_size) {
    memcpy(s, e, e->e_size);
    return e->e_size;
  } else {
    return -1;
  }
}

/****************************************************************************/
/* Does       : dump event into fd                                          */
/* Returns    : IO_OK on success, IO_WAIT if incomplete, otherwise IO_FAIL  */
/* Notes      : should involve tobuffer if we do a decent protocol format   */

int idsa_event_write(int f, IDSA_EVENT * e, int *s)
{
  int result, wr;

  if (*s >= e->e_size) {
    *s = 0;
  }
  wr = write(f, e + (*s), e->e_size - (*s));
  if (wr >= 0) {
    wr += (*s);
    if (wr == e->e_size) {
      *s = 0;
      result = IDSA_IO_OK;
    } else {
      (*s) = wr;
      result = IDSA_IO_WAIT;
    }
  } else {
    switch (errno) {
    case EAGAIN:
    case EINTR:
      result = IDSA_IO_WAIT;
      break;
    default:
      (*s) = 0;
      result = IDSA_IO_FAIL;
      break;
    }
  }
  return result;
}

/****************************************************************************/
/* Does       : copy event from buffer                                      */
/* Returns    : amount copied on success, -1 on failure                     */

int idsa_event_frombuffer(IDSA_EVENT * e, char *s, int l)
{
  if (l > sizeof(unsigned int) * 2) {
    /* 2*int <= l <= IDSA_M_MESSAGE: reasonable buffer */
    if (l > IDSA_M_MESSAGE) {
      l = IDSA_M_MESSAGE;
    }
    memcpy(e, s, (sizeof(unsigned int) * 2));

    /* IDSA_S_OFFSET <= size <= IDSA_M_MESSAGE: ok message length */
    if ((e->e_size >= IDSA_S_OFFSET) && (e->e_size <= IDSA_M_MESSAGE)) {
      if (e->e_size <= l) {	/* buffer long enough */
	memcpy(e, s, e->e_size);
	return e->e_size;
      } else {			/* buffer too short */
	return -1;
      }
    } else {			/* buggered event, block indefinitely */
      return -1;
    }
  } else {			/* buffer way too short */
    return -1;
  }
}

/****************************************************************************/
/* Does       : read event from file descriptor                             */
/* Returns    : IO_OK on success, IO_WAIT if incomplete, otherwise IO_FAIL  */
/* Notes      : will run into problems if too much data available since it  */
/*              attempts to read 4k in one go, could be fixed by first      */
/*              reading magic and size                                      */
/* Notes      : should involve frombuffer if we do a decent protocol format */

int idsa_event_read(int f, IDSA_EVENT * e, int *s)
{
  int result, rr, hv, of;

  of = (*s);
  rr = read(f, e + of, IDSA_M_MESSAGE - (*s));
  if (rr > 0) {
    hv = rr + of;
    if (hv > sizeof(unsigned int) * 2) {
      if (e->e_size == hv) {
	(*s) = 0;
	result = IDSA_IO_OK;
      } else if (e->e_size > IDSA_M_MESSAGE) {
	(*s) = 0;
	result = IDSA_IO_FAIL;
      } else {
	(*s) = hv;
	result = IDSA_IO_WAIT;
      }
    } else {
      result = IDSA_IO_WAIT;
      (*s) = hv;
    }
  } else if (rr == 0) {		/* eof */
    (*s) = 0;
    result = IDSA_IO_FAIL;
  } else {
    switch (errno) {
    case EAGAIN:
    case EINTR:
      result = IDSA_IO_WAIT;
      break;
    default:
      (*s) = 0;
      result = IDSA_IO_FAIL;
      break;
    }
  }

  return result;
}
