#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sched.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>

#include "misc.h"

#include "types.h"
#include "output.h"
#include "table.h"
#include "parse.h"
#include "idsad.h"
#include "set.h"
#include "messages.h"

#define OUTPUT_BUFFER (IDSAD_BUFFER*IDSA_M_MESSAGE)

/* functions used at startup and shutdown ********************************** */

static LOG_OUTPUT *output_new(char *s);
static int output_openfile(STATE_SET * s, char *n, int f);
static void output_openpipe(STATE_SET * s, LOG_OUTPUT * o);

static LOG_OUTPUT *output_new(char *s)
{
  LOG_OUTPUT *o;

  o = malloc(sizeof(LOG_OUTPUT));
  if (o) {
    o->o_fd = (-1);
    o->o_xd = (-1);

    o->o_type = OUTPUT_TYPE_NULL;
    o->o_flags = 0;
    o->o_format = idsa_output_lookup("native");

    strncpy(o->o_name, s, IDSA_M_MEDIUM);
    o->o_name[IDSA_M_MEDIUM - 1] = '\0';

    /* for rotate */
    o->o_sizemax = 0;
    o->o_sizehave = 0;

    /* for repeat */
    o->o_repfactor = 0;
    o->o_reptarget = 0;
    o->o_repcount = 0;
    o->o_repevent = NULL;

    /* for seekable files */
    o->o_seekback = 0;

    /* error conditions */
    o->o_dropcount = 0;
    o->o_errno = 0;

    o->o_next = NULL;
  }
  return o;
}

int output_free(LOG_OUTPUT * o)
{
  int result = 0;
  LOG_OUTPUT *p;

  while (o) {
    p = o;
    o = o->o_next;

    if (p->o_fd != (-1)) {
#ifdef TRACE
      fprintf(stderr, "output_free(): closing %d\n", p->o_fd);
#endif
      result = close(p->o_fd);
      p->o_fd = (-1);
    }
    if (p->o_xd != (-1)) {
#ifdef TRACE
      fprintf(stderr, "output_free(): closing %d\n", p->o_xd);
#endif
      result = close(p->o_xd);
      p->o_xd = (-1);
    }
    if (p->o_repevent != NULL) {
      idsa_event_free(p->o_repevent);
      p->o_repevent = NULL;
    }
    p->o_next = NULL;

    free(p);
  }

  return result;
}

/****************************************************************************/

static int output_openfile(STATE_SET * s, char *n, int f)
{
  int result;
  mode_t mode;

  mode = S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH;

  result = open(n, f, mode);
  if (result == (-1)) {
    parse_error(s, "error: unable to open file %s: %s\n", n, strerror(errno));
  } else {
    fcntl(result, F_SETFD, FD_CLOEXEC);
  }

  return result;
}

static void output_openpipe(STATE_SET * s, LOG_OUTPUT * o)
{
  int p[2];
  int status;

  if (pipe(p)) {
    parse_error(s, "error: unable to create pipe for %s: %s\n", o->o_name, strerror(errno));
  } else {
    o->o_pid = fork();
    switch (o->o_pid) {
    case -1:
      parse_error(s, "error: unable to fork %s: %s\n", o->o_name, strerror(errno));
      close(p[0]);
      close(p[1]);
      break;
    case 0:			/* in child */
#ifndef TRACE
      close(STDERR_FILENO);
#endif
      close(p[1]);
      if (dup2(p[0], STDIN_FILENO) != STDIN_FILENO) {	/* failure */
	/* WARNING: abuse of return code */
	exit(errno);
      } else {
	close(p[0]);
	strexec(o->o_name);
	/* WARNING: abuse return code again */
	exit(errno);
      }
      break;
    default:			/* in parent */
      close(p[0]);
      o->o_fd = p[1];

      /* yield to child */
      sched_yield();

      if (waitpid(o->o_pid, &status, WNOHANG)) {	/* found a child or error */
	if (WIFEXITED(status)) {
	  parse_error(s, "error: unable to start child process %s: %s\n", o->o_name, strerror(WEXITSTATUS(status)));
	} else if (WIFSIGNALED(status)) {
	  parse_error(s, "error: child process %s killed by signal %d\n", o->o_name, WTERMSIG(status));
	}
      }
      break;
    }
  }
}

void output_start(STATE_SET * s, LOG_OUTPUT * o)
{
  int flags;
  char buf[IDSA_M_MEDIUM + 3];
  struct stat sf, sx;
  int t;

  if (!(o->o_flags & OUTPUT_FLAG_ALIAS)) {

    /* this flag applies to all output types in the same way */
    if (o->o_flags & OUTPUT_FLAG_REPEAT) {	/* check for repeats */
      o->o_repevent = idsa_event_new(0);
      if (o->o_repevent == NULL) {
	parse_errormem(s, sizeof(IDSA_EVENT));
      }
    }
    /* type specific flags */
    switch (o->o_type) {
    case OUTPUT_TYPE_FILE:
      /* ATM file type can handle all options */

      flags = O_APPEND | O_CREAT | O_WRONLY;
      if (o->o_flags & OUTPUT_FLAG_SYNC) {	/* for files SYNC means open in sync mode */
	flags |= O_SYNC;
      }
      if (o->o_flags & OUTPUT_FLAG_ROTATE) {	/* flip between files */

	strncpy(buf, o->o_name, IDSA_M_MEDIUM - 1);
	buf[IDSA_M_MEDIUM] = '\0';
	strncat(buf, "-1", IDSA_M_MEDIUM + 2);
	o->o_fd = output_openfile(s, buf, flags);

	if (fstat(o->o_fd, &sf)) {
	  parse_error(s, "error: unable to stat file %s: %s\n", buf, strerror(errno));
	}
	buf[strlen(buf) - 1] = '2';
	o->o_xd = output_openfile(s, buf, flags);

	if (fstat(o->o_xd, &sx)) {
	  parse_error(s, "error: unable to stat file %s: %s\n", buf, strerror(errno));
	}
	if (parse_ok(s)) {
	  /* continue with most recent file */
	  if (sx.st_mtime > sf.st_mtime) {
	    t = o->o_xd;
	    o->o_xd = o->o_fd;
	    o->o_fd = t;

	    o->o_sizehave = sx.st_size;
	  } else {
	    o->o_sizehave = sf.st_size;
	  }
	}
      } else {			/* plain, unrotated open */
	o->o_fd = output_openfile(s, o->o_name, flags);
      }

      break;
    case OUTPUT_TYPE_PIPE:

      /* unsupported options */
      if (o->o_flags & (OUTPUT_FLAG_SEEK | OUTPUT_FLAG_ROTATE)) {
	parse_error(s, "consistency error: pipe output does not support options seek or rotate\n");
      }
      output_openpipe(s, o);

      if (!(o->o_flags & OUTPUT_FLAG_SYNC)) {	/* for pipes SYNC means don't make writes nonblocking - dangerous */
	flags = fcntl(o->o_fd, F_GETFL, 0);
	if (flags != (-1)) {
	  fcntl(o->o_fd, F_SETFL, O_NONBLOCK | flags);
	} else {
	  parse_error(s, "error: unable to make pipe %d nonblocking: %s\n", o->o_fd, strerror(errno));
	}
      }
      break;
    default:
      parse_error(s, "consistency error: unknown output type 0x%02x\n", o->o_type);
      break;
    }
  }
}

LOG_OUTPUT *output_insert(LOG_OUTPUT * o, char *s)
{
  /* return the new / inserted one at start of list */
  LOG_OUTPUT *p, *q;

  /* look for a match, get the one in front of it too */
  p = NULL;
  q = o;
  while (q && strncmp(s, q->o_name, IDSA_M_MEDIUM)) {
    p = q;
    q = q->o_next;
  }

  if (q == NULL) {		/* no match, allocate */
    q = output_new(s);
    if (q) {			/* prepend new element */
      q->o_next = o;
      o = q;
    } else {			/* ouch, better delete the lot */
      output_free(o);
      o = NULL;
    }
  } else {			/* match, shuffle it to the front */
#ifdef TRACE
    fprintf(stderr, "output_insert(): found match for %s\n", s);
#endif
    if (p) {			/* not at start of list, so work to do */
      p->o_next = q->o_next;
      q->o_next = o;
      o = q;
    }
    q->o_flags |= OUTPUT_FLAG_ALIAS;
  }

  return o;
}

void output_dump(LOG_OUTPUT * o, FILE * f)
{
  while (o) {
    fprintf(f, "output: type 0x%02x, flags 0x%02x, %s\n", o->o_type, o->o_flags, o->o_name);
    fprintf(f, "output: size: max %d, have %d\n", o->o_sizemax, o->o_sizehave);
    fprintf(f, "output: repeat: factor %d, target %d, count %d\n", o->o_repfactor, o->o_reptarget, o->o_repcount);
    o = o->o_next;
  }
}

/* runtime output functions ************************************************ */

static int output_isrepeat(IDSA_EVENT * p, IDSA_EVENT * n);
static int output_write(LOG_OUTPUT * o, IDSA_EVENT * e);

static int output_isrepeat(IDSA_EVENT * p, IDSA_EVENT * n)
{
  IDSA_UNIT *u, *v;

  if (p->e_size != n->e_size)
    return 0;

  /* ick, time wants special treatment */
  u = idsa_event_unitbynumber(p, IDSA_N_TIME);
  v = idsa_event_unitbynumber(n, IDSA_N_TIME);
  if (v && u) {
    idsa_unit_copy(u, v);
  }
  if (memcmp(p->e_ptr, n->e_ptr, n->e_size - IDSA_S_OFFSET))
    return 0;

#ifdef TRACE
  fprintf(stderr, "output_isrepeat(): are repeated\n");
#endif

  return 1;
}

static int output_write(LOG_OUTPUT * o, IDSA_EVENT * e)
{
  int result = 0;
  int wr, sw;
  char buffer[OUTPUT_BUFFER];

#ifdef TRACE
  fprintf(stderr, "output_write(): should write to %d\n", o->o_fd);
#endif

  sw = idsa_output(e, buffer, OUTPUT_BUFFER, o->o_format);

  if (sw > 0) {			/* something to write */
    wr = write(o->o_fd, buffer, sw);
    if (wr != sw) {
      if (wr > 0) {		/* partial write, yeech */
	if (o->o_flags & OUTPUT_FLAG_SEEK) {	/* try to rewind */
	  lseek(o->o_fd, (-1) * wr, SEEK_END);
	}
	/* should I assign some value to o->o_errno ? */
      } else if (wr < 0) {	/* write failure, write sets errno */
	o->o_errno = errno;
      }
      o->o_dropcount++;
      result++;
      o->o_seekback = 0;
    } else {
      o->o_seekback = wr;
    }
  } else {			/* nothing to write */
    if (sw < 0) {		/* and it is a failure, ouch */
      /* should I assign some value to o->o_errno ? */
      o->o_dropcount++;
      result++;
    }
    o->o_seekback = 0;
  }

  return result;
}

int output_run(LOG_OUTPUT * o, IDSA_EVENT * e)
{
  int result = 0;
  int t;

#ifdef TRACE
  fprintf(stderr, "output_run(): should log event\n");
#endif

  if (o->o_repfactor) {		/* need to check out repeats */
    if (output_isrepeat(o->o_repevent, e)) {	/* yup, a repeat */
#ifdef TRACE
      fprintf(stderr, "output_run(): event is repeated\n");
#endif
      if (o->o_repcount >= o->o_reptarget) {	/* flush if we have hit limit */
#ifdef TRACE
	fprintf(stderr, "output_run(): flushing since count=%d >= target=%d\n", o->o_repcount, o->o_reptarget);
#endif
	o->o_repevent->e_repeat = o->o_repcount;
	result += output_write(o, o->o_repevent);
	/* FIXME: update time */
	o->o_repcount = 0;
	o->o_reptarget = o->o_reptarget * o->o_repfactor;
      } else {			/* don't write yet, just bump counter */
	o->o_repcount++;
#ifdef TRACE
	fprintf(stderr, "output_run(): No write, just bumping repeat counter to %d\n", o->o_repcount);
#endif
      }
    } else {			/* not a repeated message */
      if (o->o_repcount) {	/* previous repeat needs to be flushed */
#ifdef TRACE
	fprintf(stderr, "output_run(): flushing last repeat\n");
#endif
	o->o_repevent->e_repeat = o->o_repcount;
	result += output_write(o, o->o_repevent);
	o->o_repcount = 0;
	/* FIXME: maybe have it decay ? */
	o->o_reptarget = 1;
      }
#ifdef TRACE
      fprintf(stderr, "output_run(): writing new message\n");
#endif
      result += output_write(o, e);
      idsa_event_copy(o->o_repevent, e);
    }
  } else {			/* could not be bothered about repeats */
#ifdef TRACE
    fprintf(stderr, "output_run(): plain log, no repeat check\n");
#endif
    result += output_write(o, e);
  }

#ifdef TRACE
  fprintf(stderr, "output_run(): have %d, max %d\n", o->o_sizehave, o->o_sizemax);
#endif
  /* rotate logs if required */
  if ((o->o_sizemax > 0) && (o->o_sizemax <= o->o_sizehave)) {
#ifdef TRACE
    fprintf(stderr, "output_run(): rotating\n");
#endif
    t = o->o_fd;
    o->o_fd = o->o_xd;
    o->o_xd = t;

    if (ftruncate(o->o_fd, 0)) {
      o->o_errno = errno;
      result++;
    }
    o->o_sizehave = 0;
  }
  if (!(o->o_flags & OUTPUT_FLAG_REPORT)) {	/* if report not set then ignore failures */
    result = 0;
  }
  return result;
}

void output_error(STATE_SET * s)
{
  LOG_OUTPUT *o;

  if (s->s_outputfail) {
    o = s->s_output;
    while (o) {
      if (o->o_dropcount && (o->o_flags & OUTPUT_FLAG_REPORT)) {
	message_error_output(s, o);
	o->o_dropcount = 0;
      }
      o = o->o_next;
    }
    s->s_outputfail = 0;
  }
}
