#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <pwd.h>

#include <idsa_internal.h>

#include "mex.h"

#include "idsad.h"
#include "rule.h"
#include "parse.h"
#include "output.h"
#include "set.h"
#include "branch.h"
#include "table.h"
#include "queue.h"
#include "mod.h"

static int parse_stop(MEX * l);
static MEX *parse_start(char *s);

static RULE_BODY *parse_rule(STATE_SET * s, MEX * l);

static RULE_MATCH *parse_match(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_clause(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_term(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_open(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_not(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_mod(STATE_SET * s, MEX * l);
static RULE_MATCH *parse_element(STATE_SET * s, MEX * l, char *n);

static void parse_actions(STATE_SET * s, RULE_BODY * r, MEX * l);
static RULE_ACTION *parse_send(STATE_SET * s, MEX * l);
static RULE_ACTION *parse_keep(STATE_SET * s, MEX * l);
static void parse_keepoptions(STATE_SET * s, QUEUE_KEEP * k, MEX * l);
static RULE_ACTION *parse_log(STATE_SET * s, RULE_BODY * r, MEX * l);
static void parse_logoptions(STATE_SET * s, LOG_OUTPUT * o, MEX * l);

static unsigned int parse_type(STATE_SET * s, MEX * l, char *n);

/* reserved characters ***************************************************** */

struct parse_characters {
  int p_id;
  unsigned char p_char;
};

#define PARSE_COLON           0x01	/* : */
#define PARSE_SCOLON          0x02	/* ; */
#define PARSE_OPEN            0x03	/* ( */
#define PARSE_CLOSE           0x04	/* ) */
#define PARSE_NOT             0x05	/* ! */
#define PARSE_AND             0x06	/* & */
#define PARSE_OR              0x07	/* | */
#define PARSE_COMMA           0x08	/* , */
#define PARSE_DOLLAR          0x09	/* $ */
#define PARSE_EQUAL           0x0a	/* = */
#define PARSE_MOD              0x0a	/* % */

struct parse_characters pc_table[] =
{
  {PARSE_COLON, ':'},
  {PARSE_SCOLON, ';'},
  {PARSE_OPEN, '('},
  {PARSE_CLOSE, ')'},
  {PARSE_NOT, '!'},
  {PARSE_AND, '&'},
  {PARSE_OR, '|'},
  {PARSE_COMMA, ','},
  {PARSE_DOLLAR, '$'},
  {PARSE_EQUAL, '='},
  {PARSE_MOD, '%'},
  {0, '\0'}
};

/* reserverd words ********************************************************* */

struct parse_words {
  int p_id;
  char *p_word;
};

#define PARSE_ALLOW           0x10	/* allow */
#define PARSE_DENY            0x11	/* deny */
#define PARSE_CONTINUE        0x12	/* continue */
#define PARSE_DROP            0x13	/* drop */
#define PARSE_SEND            0x14	/* send */

#define PARSE_KEEP            0x20	/* keep */
#define PARSE_SIZE            0x21	/* number of items kept */
#define PARSE_TIMEOUT         0x22	/* length for which it is kept */

#define PARSE_LOG             0x30	/* log */
#define PARSE_FILE            0x31	/* to file */
#define PARSE_PIPE            0x32	/* to pipe */

#define PARSE_SYNC            0x33	/* file needs to be flushed */
#define PARSE_SEEK            0x34	/* file is seekable */
#define PARSE_REPEAT          0x35	/* print repeat count for duplicates */
#define PARSE_ROTATE          0x36	/* rotate file */
#define PARSE_REPORT          0x37	/* log failures writing to this target */

#define PARSE_FAIL            0x38	/* failure of module */

#define PARSE_FORMAT          0x40	/* format */

#define PARSE_ARISK           0x50	/* risk:availability */
#define PARSE_CRISK           0x51	/* risk:confidentiality */
#define PARSE_IRISK           0x52	/* risk:integrity */

struct parse_words pw_table[] =
{
  {PARSE_ALLOW, "allow"},
  {PARSE_DENY, "deny"},
  {PARSE_CONTINUE, "continue"},
  {PARSE_DROP, "drop"},
  {PARSE_SEND, "send"},
  {PARSE_KEEP, "keep"},
  {PARSE_SIZE, "size"},
  {PARSE_TIMEOUT, "timeout"},
  {PARSE_LOG, "log"},
  {PARSE_FILE, "file"},
  {PARSE_PIPE, "pipe"},
  {PARSE_SYNC, "sync"},
  {PARSE_SEEK, "seek"},
  {PARSE_REPEAT, "repeat"},
  {PARSE_ROTATE, "rotate"},
  {PARSE_REPORT, "report"},
  {PARSE_FORMAT, "format"},
  {PARSE_FAIL, "fail"},
  {PARSE_ARISK, "arisk"},
  {PARSE_CRISK, "crisk"},
  {PARSE_IRISK, "irisk"},
  {0, "\0"}
};

static MEX *parse_start(char *s)
{
  MEX *l;
  int i;

  l = mex_open(s);
  if (l) {
    for (i = 0; pc_table[i].p_id != 0; i++) {
      mex_keychar(l, pc_table[i].p_char, pc_table[i].p_id);
    }

    for (i = 0; pw_table[i].p_id != 0; i++) {
      mex_keyword(l, pw_table[i].p_word, pw_table[i].p_id);
    }
  }
#ifdef TRACE
  fprintf(stderr, "parse_start(): set up key characters and words\n");
#endif

  return l;
}

static int parse_stop(MEX * l)
{
#ifdef TRACE
  fprintf(stderr, "parse_stop(): stopping\n");
#endif
  return mex_close(l);
}

STATE_SET *parse_set(char *n)
{
  MEX *l;
  MEX_TOKEN *t;
  RULE_BODY *r, *q;
  STATE_SET *s;

  s = set_new();
  if (s) {
    l = parse_start(n);		/* boot tokenizer */
    if (l) {

#ifdef TRACE
      mex_dump(l, stderr);
#endif

      q = NULL;
      do {			/* read in rules */
#ifdef TRACE
	fprintf(stderr, "parse_set(): grabbing new rule\n");
#endif
	r = parse_rule(s, l);
	if (r) {
#ifdef TRACE
	  rule_dump(r, stderr);
#endif
	  if (q) {
	    q->r_next = r;
	  } else {
	    s->s_rules = r;
	  }
	  q = r;
	}
      } while (r);

#ifdef TRACE
      fprintf(stderr, "parse_set(): finished parsing\n");
#endif

      /* check for failure */
      if (l->m_error) {
	parse_error(s, "tokenizing error: %s", mex_error(l));
      } else {			/* not reading the entire file is an error */
	t = mex_get(l);
	if (t) {
	  parse_error(s, "parse error on line %d: parsing aborted before token %s", t->t_line, t->t_buf);
	}
      }

      parse_stop(l);		/* delete tokenizer */
    } else {
      parse_error(s, "tokenizing error: unable to start tokenizing file %s", n);
    }
  }
  return s;
}

static RULE_BODY *parse_rule(STATE_SET * s, MEX * l)
{
  RULE_BODY *r = NULL;
  RULE_MATCH *m;
  MEX_TOKEN *t;

  r = rule_new();
  if (r) {
    s->s_currentrule = r;	/* transient, used to set h_rule */

    s->s_rulecount++;
    r->r_number = s->s_rulecount;

#ifdef TRACE
    fprintf(stderr, "parse_rule(): starting on rule %d\n", r->r_number);
#endif
    m = parse_match(s, l);
    if (m) {
      r->r_match = m;
      t = mex_get(l);
      if (t) {
	if (t->t_id == PARSE_COLON) {
#ifdef TRACE
	  fprintf(stderr, "parse_rule(): found : now parsing actions\n");
#endif
	  parse_actions(s, r, l);
	} else {
	  parse_errortoken(s, t);
	}
      } else {
	parse_erroreof(s);
      }
    }
    if ((m == NULL) || (!parse_ok(s))) {
#ifdef TRACE
      fprintf(stderr, "parse_rule(): failure in rule %d\n", s->s_rulecount);
#endif
      rule_free(r);
      r = NULL;
    }
  } else {
    parse_error(s, "unable to allocate memory for new rule");
  }

  return r;
}

static RULE_MATCH *parse_match(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m, *n;
  MEX_TOKEN *t;

  m = parse_clause(s, l);
  if (m) {
    t = mex_get(l);
    if (t) {
      if (t->t_id == PARSE_OR) {
#ifdef TRACE
	fprintf(stderr, "parse_match(): found or, looking at right side\n");
#endif
	n = match_new(RULE_GATE_OR);
	if (n == NULL) {
	  parse_errormem(s, sizeof(RULE_MATCH));
	} else {
	  n->m_left = m;
	  m->m_up = n;
	  n->m_right = parse_match(s, l);
	  if (n->m_right) {
	    n->m_right->m_up = n;
	  } else {
	    parse_errortoken(s, t);
	  }
	  m = n;
	}
      } else {
	mex_unget(l, t);
      }
    }
  }
  return m;
}

static RULE_MATCH *parse_clause(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m, *n;
  MEX_TOKEN *t;

  m = parse_term(s, l);
  if (m) {
    t = mex_get(l);
    if (t) {
      if (t->t_id == PARSE_AND) {
#ifdef TRACE
	fprintf(stderr, "parse_clause(): found and, looking at right side\n");
#endif
	n = match_new(RULE_GATE_AND);
	if (n == NULL) {
	  parse_errormem(s, sizeof(RULE_MATCH));
	} else {
	  n->m_left = m;
	  m->m_up = n;
	  n->m_right = parse_clause(s, l);
	  if (n->m_right) {
	    n->m_right->m_up = n;
	  } else {
	    parse_errortoken(s, t);
	  }
	  m = n;
	}
      } else {
	mex_unget(l, t);
      }
    }
  }
  return m;
}

static RULE_MATCH *parse_open(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m;
  MEX_TOKEN *t;

#ifdef TRACE
  fprintf(stderr, "parse_open(): found (, restarting\n");
#endif

  m = parse_match(s, l);
  if (m) {
    t = mex_get(l);
    if (t) {
      if (t->t_id != PARSE_CLOSE) {
	parse_error(s, "parse error on line %d: expecting ) instead of %s", t->t_line, t->t_buf);
      }
    } else {
      parse_error(s, "parse error: unmatched (");
    }
  } else {
    parse_erroreof(s);
  }

  return m;
}

static RULE_MATCH *parse_not(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m;

#ifdef TRACE
  fprintf(stderr, "parse_not(): found not, negating\n");
#endif

  m = parse_term(s, l);
  if (m) {
    if (m->m_gate & RULE_GATE_NEGATE) {		/* double negative */
      m->m_gate &= (~RULE_GATE_NEGATE);
    } else {
      m->m_gate |= RULE_GATE_NEGATE;
    }
  } else {
    parse_errorrule(s);
  }

  return m;
}

static RULE_MATCH *parse_term(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m;
  MEX_TOKEN *t;

  m = NULL;
  t = mex_get(l);
  if (t) {
    switch (t->t_id) {
    case PARSE_OPEN:		/* ( ... */
      m = parse_open(s, l);
      break;
    case PARSE_NOT:		/* ~ ... */
      m = parse_not(s, l);
      break;
    case PARSE_MOD:		/* ! ... */
      m = parse_mod(s, l);
      break;
    default:
      m = parse_element(s, l, t->t_buf);
      break;
    }
  }				/* else return NULL which is not always an error */
  return m;
}

static RULE_MATCH *parse_mod(STATE_SET * s, MEX * l)
{
  RULE_MATCH *m;
  MEX_TOKEN *t, *u;
  int failmode;
  int timeout;

  failmode = MOD_FAILCLOSED;
  timeout = IDSAD_MODTIMEOUT;

  m = NULL;
  t = mex_get(l);

  if (t) {

    /* parse any options */
    u = mex_get(l);
    while (u) {
      if (u->t_id == PARSE_COMMA) {	/* options all start with a comma */
	u = mex_get(l);
	if (u) {
	  switch (u->t_id) {
	  case PARSE_FAIL:	/* fail allow, fail deny */
	    u = mex_get(l);
	    if (u) {
	      switch (u->t_id) {
	      case PARSE_DENY:
		failmode = MOD_FAILCLOSED;
		u = mex_get(l);
		break;
	      case PARSE_ALLOW:
		failmode = MOD_FAILOPEN;
		u = mex_get(l);
		break;
	      default:
		parse_error(s, "parse error on line %d: unknown failure mode %s", u->t_line, u->t_buf);
		u = NULL;
		break;
	      }
	    } else {
	      parse_erroreof(s);
	    }
	    break;
	  case PARSE_TIMEOUT:	/* timeout integer */
	    u = mex_get(l);
	    if (u) {
	      timeout = atoi(u->t_buf);
	      if (timeout) {
		u = mex_get(l);
	      } else {
		parse_error(s, "parse error on line %d: %s not a valid timeout", u->t_line, u->t_buf);
		u = NULL;
	      }
	    } else {
	      parse_erroreof(s);
	    }
	    break;
	  default:		/* unknown option */
	    parse_error(s, "parse error on line %d: unknown %% option %s", u->t_line, u->t_buf);
	    u = NULL;
	    break;
	  }
	} else {
	  parse_erroreof(s);
	}
      } else {			/* not a comma, not an option, unget */
	mex_unget(l, u);
	u = NULL;
      }
    }

    if (s->s_errno == 0) {	/* no errors in retrieving options */
      m = mod_match_make(s, t->t_buf, timeout, failmode);
    }
  } else {
    parse_erroreof(s);
  }

  return m;
}

static RULE_MATCH *parse_element(STATE_SET * s, MEX * l, char *n)
{
  MEX_TOKEN *t;
  RULE_MATCH *m;
  unsigned int y;

#ifdef TRACE
  fprintf(stderr, "parse_element(): parsing <%s>\n", n);
#endif

  m = NULL;
  y = parse_type(s, l, n);

#ifdef TRACE
  fprintf(stderr, "parse_element(): provisional type is 0x%04x\n", y);
#endif

  t = mex_get(l);
  if (t) {
    if (t->t_id == PARSE_DOLLAR) {
      t = mex_get(l);
      if (t) {
#ifdef TRACE
	fprintf(stderr, "parse_element(): inserting queue entry\n");
#endif
	m = queue_match_make(s, t->t_buf, y, n);
      }
    } else {
#ifdef TRACE
      fprintf(stderr, "parse_element(): inserting table entry\n");
#endif
      m = table_match_make(s, y, n, t->t_buf);
    }
  } else {
    parse_erroreof(s);
  }

  return m;
}

static void parse_actions(STATE_SET * s, RULE_BODY * r, MEX * l)
{
  /* errors get passed via error using parse_error() */
  RULE_ACTION *a, *b;
  int run;
  MEX_TOKEN *t;

  run = 1;
  a = NULL;
  b = NULL;
  t = mex_get(l);
  if (t) {
    do {
#ifdef TRACE
      fprintf(stderr, "parse_actions(): found action %s:%d\n", t->t_buf, t->t_id);
#endif
      switch (t->t_id) {
      case PARSE_ALLOW:
	r->r_policy |= RULE_ACTION_ALLOW;
	break;
      case PARSE_DENY:
	r->r_policy |= RULE_ACTION_DENY;
	break;
      case PARSE_CONTINUE:
	r->r_policy |= RULE_ACTION_CONTINUE;
	break;
      case PARSE_DROP:
	r->r_policy |= RULE_ACTION_DROP;
	break;
      case PARSE_KEEP:
	a = parse_keep(s, l);
	if (a) {
	  if (b) {
	    b->a_next = a;
	  } else {
	    r->r_action = a;
	  }
	  b = a;
	} else {
	  run = 0;
	}
	r->r_policy |= RULE_ACTION_SEND;
	break;
      case PARSE_SEND:
	a = parse_send(s, l);
	if (a) {
	  if (b) {
	    b->a_next = a;
	  } else {
	    r->r_action = a;
	  }
	  b = a;
	} else {
	  run = 0;
	}
	r->r_policy |= RULE_ACTION_SEND;
	break;
      case PARSE_LOG:
	a = parse_log(s, r, l);
	if (a) {
	  if (b) {
	    b->a_next = a;
	  } else {
	    r->r_action = a;
	  }
	  b = a;
	} else {
	  run = 0;
	}
	r->r_policy |= RULE_ACTION_LOG;
	break;
      default:
	parse_error(s, "parse error on line %d: unknown action %s", t->t_line, t->t_buf);
	run = 0;
	break;
      }

      t = mex_get(l);
      if (t) {
	if (t->t_id == PARSE_SCOLON) {
	  t = mex_get(l);
	  if (t == NULL) {
	    parse_error(s, "parse error: incomplete action in rule %d", r->r_number);
	    run = 0;
	  }
	} else {
	  mex_unget(l, t);
	  run = 0;
	}
      } else {
	run = 0;
      }
    } while (run);
  } else {
    parse_error(s, "parse error: rule %d has no action", r->r_number);
  }

  /* note - default policy is to allow things */
  if (!(r->r_policy & RULE_ACTION_DENY)) {
    r->r_policy = r->r_policy | RULE_ACTION_ALLOW;
  }
}

static unsigned int parse_type(STATE_SET * s, MEX * l, char *n)
{
  MEX_TOKEN *t;
  unsigned int result = IDSA_T_NULL;
  unsigned int lookup;

  lookup = idsa_reserved_typebyname(n);

  t = mex_get(l);
  if (t) {
    if (t->t_id == PARSE_EQUAL) {
      t = mex_get(l);
      if (t) {
	result = idsa_unit_type(t->t_buf);
#ifdef TRACE
	fprintf(stderr, "parse_type(): have equal, type of <%s> is 0x%04x\n", t->t_buf, result);
#endif
      } else {
	parse_erroreof(s);
      }
    } else {
#ifdef TRACE
      fprintf(stderr, "parse_type(): <%s> not =, ungetting\n", t->t_buf);
#endif
      mex_unget(l, t);
    }
  } else {
    parse_erroreof(s);
  }

  if ((lookup == IDSA_T_NULL) && (result == IDSA_T_NULL)) {
  } else if ((result != lookup) && (lookup != IDSA_T_NULL) && (result != IDSA_T_NULL)) {
    parse_error(s, "type conflict for %s: reserved type is %s", n, idsa_unit_name(lookup));
  } else if (result == IDSA_T_NULL) {
    result = lookup;
  }
  return result;
}

static void parse_keepoptions(STATE_SET * s, QUEUE_KEEP * k, MEX * l)
{
  MEX_TOKEN *tc, *to, *ts;
  int run = 1;
  do {
    tc = mex_get(l);
    if (tc && (tc->t_id == PARSE_COMMA)) {
      to = mex_get(l);
      ts = mex_get(l);
      if (to && ts) {
	switch (to->t_id) {
	case PARSE_TIMEOUT:
	  queue_set_timeout(s, k, atoi(ts->t_buf));
	  break;
	case PARSE_SIZE:
	  queue_set_size(s, k, atoi(ts->t_buf));
	  break;
	default:
	  parse_error(s, "parse error: unknown option %s for keep", to->t_buf);
	  break;
	}
      } else {
	parse_erroreof(s);
	run = 0;
      }
    } else {
      if (tc)
	mex_unget(l, tc);
      run = 0;
    }
  } while (run);
}

static RULE_ACTION *parse_keep(STATE_SET * s, MEX * l)
{
  RULE_ACTION *a;
  MEX_TOKEN *tn, *td, *tv;
  QUEUE_KEEP *k;
  unsigned int y;

  a = NULL;

  tn = mex_get(l);
  if (tn) {
#ifdef TRACE
    fprintf(stderr, "parse_keep(): keeping %s\n", tn->t_buf);
#endif
    y = parse_type(s, l, tn->t_buf);
    td = mex_get(l);
    if (td) {
      if (td->t_id == PARSE_DOLLAR) {
	tv = mex_get(l);
	if (tv) {
	  k = queue_keep_make(s, tv->t_buf, y, tn->t_buf);
	  if (k) {
	    queue_set_size(s, k, 1);	/* note 1!=l */
	    parse_keepoptions(s, k, l);

	    a = action_new(RULE_ACTION_KEEP);
	    if (a) {
	      a->a_un.a_keep = k;
	    } else {
	      queue_keep_free(k);
	      parse_errormem(s, sizeof(RULE_ACTION));
	    }

	  }
	} else {
	  parse_erroreof(s);
	}
      } else {
	parse_error(s, "parse error: expected $ after keep %s", tn->t_buf);
      }
    } else {
      parse_erroreof(s);
    }
  } else {
    parse_erroreof(s);
  }

  return a;
}

static RULE_ACTION *parse_send(STATE_SET * s, MEX * l)
{
  RULE_ACTION *a;
  MEX_TOKEN *t, *n;
  IDSA_UNIT *u;
  unsigned int y;

  a = NULL;
  n = mex_get(l);
  if (n) {
    y = parse_type(s, l, n->t_buf);
    if (y != IDSA_T_NULL) {
      t = mex_get(l);
      if (t) {
	u = idsa_unit_new(n->t_buf, y, t->t_buf);
	if (u) {
	  a = action_new(RULE_ACTION_SEND);
	  if (a) {
	    a->a_un.a_unit = u;
	  } else {
	    parse_errormem(s, sizeof(RULE_ACTION));
	  }
	}
      } else {
	parse_erroreof(s);
      }
    } else {
      parse_error(s, "require a type for %s", n->t_buf);
    }
  } else {
    parse_erroreof(s);
  }

  return a;
}

static void parse_logoptions(STATE_SET * s, LOG_OUTPUT * o, MEX * l)
{
  MEX_TOKEN *t;

  do {
    t = mex_get(l);
    if (t) {
      if (t->t_id == PARSE_COMMA) {
	if (o->o_flags & OUTPUT_FLAG_ALIAS) {
	  parse_error(s, "parse error on line %d: log options for %s may only be specified once", t->t_line, o->o_name);
	  t = NULL;
	} else {
	  t = mex_get(l);
	  if (t) {
#ifdef TRACE
	    fprintf(stderr, "parse_logoptions(): looking at option %s\n", t->t_buf);
#endif
	    switch (t->t_id) {
	    case PARSE_SEEK:
	      o->o_flags |= OUTPUT_FLAG_SEEK;
	      break;
	    case PARSE_SYNC:
	      o->o_flags |= OUTPUT_FLAG_SYNC;
	      break;
	    case PARSE_REPORT:
	      o->o_flags |= OUTPUT_FLAG_REPORT;
	      break;
	    case PARSE_REPEAT:
	      o->o_flags |= OUTPUT_FLAG_REPEAT;
	      t = mex_get(l);
	      if (t) {
		o->o_reptarget = 1;
		o->o_repfactor = atoi(t->t_buf);
		if (o->o_repfactor <= 1) {
		  parse_errorinvalid(s, t, "repeat factor");
		}
	      } else {
		parse_erroreof(s);
	      }
	      break;
	    case PARSE_ROTATE:
	      o->o_flags |= OUTPUT_FLAG_ROTATE;
	      t = mex_get(l);
	      if (t) {
		o->o_sizemax = atoi(t->t_buf);
		if (o->o_sizemax <= sizeof(IDSA_EVENT)) {
		  parse_errorinvalid(s, t, "log size");
		}
	      } else {
		parse_erroreof(s);
	      }
	      break;
	    case PARSE_FORMAT:
	      t = mex_get(l);
	      if (t) {
		o->o_format = idsa_output_lookup(t->t_buf);
		if (o->o_format == NULL) {
		  parse_errorinvalid(s, t, "output format");
		  break;
		}
	      } else {
		parse_erroreof(s);
	      }
	      break;
	    default:
	      parse_errorinvalid(s, t, "log option");
	      t = NULL;
	      break;
	    }
	  } else {
	    parse_erroreof(s);
	  }
	}
      } else {
	mex_unget(l, t);
	t = NULL;
      }
    }
  } while (t);
}

static RULE_ACTION *parse_log(STATE_SET * s, RULE_BODY * r, MEX * l)
{
  RULE_ACTION *a = NULL;
  MEX_TOKEN *t;
  int type;
  char *format;

  format = "native";
  t = mex_get(l);
  if (t) {
    switch (t->t_id) {		/* what type of target - file, pipe */
    case PARSE_FILE:
      type = OUTPUT_TYPE_FILE;
      format = "native";
      break;
    case PARSE_PIPE:
      type = OUTPUT_TYPE_PIPE;
      format = "internal";
      break;
    default:
      parse_errorinvalid(s, t, "log type");
      type = OUTPUT_TYPE_NULL;
      break;
    }

    t = mex_get(l);
    if (t) {
#ifdef TRACE
      fprintf(stderr, "parse_log(): log destination is <%s>\n", t->t_buf);
#endif
      if (strlen(t->t_buf) >= IDSA_M_MEDIUM) {
	parse_error(s, "parse error on line %d: %s is too long", t->t_line, t->t_buf);
      } else {
	s->s_output = output_insert(s->s_output, t->t_buf);
	if (s->s_output) {
	  if (!(s->s_output->o_flags & OUTPUT_FLAG_ALIAS)) {
	    s->s_output->o_format = idsa_output_lookup(format);
	    if (s->s_output->o_format == NULL) {
	      parse_errorinvalid(s, t, "default log format");
	    }
	    s->s_output->o_type = type;
	  }
	  a = action_new(RULE_ACTION_LOG);
	  if (a) {
	    a->a_un.a_output = s->s_output;
	    parse_logoptions(s, s->s_output, l);
#ifdef TRACE
	    output_dump(s->s_output, stderr);
#endif
	    output_start(s, s->s_output);
	  } else {
	    parse_errormem(s, sizeof(RULE_ACTION));
	  }
	} else {
	  parse_errormem(s, sizeof(LOG_OUTPUT));
	}
      }

    } else {
      parse_erroreof(s);
    }
  } else {
    parse_erroreof(s);
  }

  return a;
}

int parse_ok(STATE_SET * s)
{
  return (s->s_errno ? 0 : 1);
}

void parse_error(STATE_SET * s, char *e,...)
{
  va_list ap;

  va_start(ap, e);
  if (s->s_errno == 0) {

#ifdef TRACE
    fputs(s->s_errbuf, stderr);
#endif

    vsnprintf(s->s_errbuf, IDSA_M_LONG, e, ap);
    s->s_errbuf[IDSA_M_LONG - 1] = '\0';
    s->s_errno = 1;
  }
  va_end(ap);
}

void parse_errortoken(STATE_SET * s, MEX_TOKEN * t)
{
  parse_error(s, "parse error on line %d near token %s", t->t_line, t->t_buf);
}

void parse_errorrule(STATE_SET * s)
{
  parse_error(s, "parse error in rule number %d", s->s_rulecount);
}

void parse_erroreof(STATE_SET * s)
{
  parse_error(s, "premature eof while parsing rule number %d", s->s_rulecount);
}

void parse_errormem(STATE_SET * s, int b)
{
  parse_error(s, "memory allocation failure: unable to allocate %d bytes", b);
}

void parse_errorinvalid(STATE_SET * s, MEX_TOKEN * t, char *m)
{
  parse_error(s, "parse error on line %d: %s is not a valid %s", t->t_line, t->t_buf, m);
}
